@@ -9,8 +9,7 @@
 #' @param testData The input test dataset. The first column
 #' is the label or the output. For binary classes,
 #' 0 and 1 are used to indicate the class member.
-#' @param predMode The prediction mode. Available options are
-#' c('probability', 'classification').
+#' @param predMode The prediction mode.Currently only supports 'probability' for binary classification tasks.
 #' @param classifier  Learners in mlr3
 #' @param paramlist   Learner parameters
 #' @param inner_folds k-fold cross validation ( Only supported when testData = NULL )
@@ -153,6 +152,8 @@ baseModel=function ( trainData, testData, predMode = "probability",
 #' @param pathlistDB_sub A list of pathways with pathway IDs and their
 #' corresponding genes ('entrezID' is used).
 #' For details, please refer to ( data("GO2ALLEGS_BP") )
+#' @param MinfeatureNum_pathways The minimal defined pathway size after mapping your
+#' own data to pathlistDB(KEGG database/GO database).
 #' @param cores The number of cores used for computation.
 #' @param verbose Whether to print running process information to the console
 #'
@@ -171,7 +172,7 @@ baseModel=function ( trainData, testData, predMode = "probability",
 #'                  featureAnno=MethylAnno,pathlistDB_sub=GO2ALLEGS_BP,cores=1)
 #'
 Stage1_FeartureSelection=function(Stage1_FeartureSelection_Method='cor',data=NULL,cutoff=NULL,
-                                 featureAnno=NULL,pathlistDB_sub=NULL,cores=1,verbose=TRUE){
+                                 featureAnno=NULL,pathlistDB_sub=NULL,MinfeatureNum_pathways=10,cores=1,verbose=TRUE){
   if(Sys.info()[1]=="Windows"){
     cores=1
   }
@@ -184,13 +185,14 @@ Stage1_FeartureSelection=function(Stage1_FeartureSelection_Method='cor',data=NUL
     Cor_names=names(Cor)
     Cor_cutoff=Cor[which(Cor>cutoff)]
     Cor_cutoff_names=names(Cor_cutoff)
+    MinfeatureNum_pathways2=MinfeatureNum_pathways+1
     feature_pathways=mclapply(1:length(pathlistDB_sub),function(x){
       id=c('label',featureAnno$ID[which(featureAnno$entrezID %in% pathlistDB_sub[[x]])])
-      if(length(id)>10){
+      if(length(id)>MinfeatureNum_pathways){
         id2=id[which(id %in% Cor_cutoff_names)]
-        if(length(id2)<11){
+        if(length(id2)<MinfeatureNum_pathways2){
           a=Cor[id]
-          id2=names(a)[order(a,decreasing = T)[1:11]]
+          id2=names(a)[order(a,decreasing = T)[1:MinfeatureNum_pathways2]]
           return(id2)
         }else{
           return(id2)
@@ -211,13 +213,14 @@ Stage1_FeartureSelection=function(Stage1_FeartureSelection_Method='cor',data=NUL
     Cor_names=names(Cor)
     Cor_cutoff=Cor[which(Cor<cutoff)]
     Cor_cutoff_names=names(Cor_cutoff)
+    MinfeatureNum_pathways2= MinfeatureNum_pathways+1
     feature_pathways=mclapply(1:length(pathlistDB_sub),function(x){
       id=c('label',featureAnno$ID[which(featureAnno$entrezID %in% pathlistDB_sub[[x]])])
-      if(length(id)>10){
+      if(length(id)> MinfeatureNum_pathways){
         id2=id[which(id %in% Cor_cutoff_names)]
-        if(length(id2)<11){
+        if(length(id2)< MinfeatureNum_pathways2){
           a=Cor[id]
-          id2=names(a)[order(a,decreasing = T)[1:11]]
+          id2=names(a)[order(a,decreasing = F)[1:MinfeatureNum_pathways2]]
           return(id2)
         }else{
           return(id2)
@@ -271,7 +274,7 @@ Stage1_FeartureSelection=function(Stage1_FeartureSelection_Method='cor',data=NUL
         id2=id[which(id %in% Cor_cutoff_names)]
         if(length(id2)<11){
           a=Cor[id]
-          id2=names(a)[order(a,decreasing = T)[1:11]]
+          id2=names(a)[order(a,decreasing = F)[1:11]]
           return(id2)
         }else{
           return(id2)
@@ -304,8 +307,7 @@ Stage1_FeartureSelection=function(Stage1_FeartureSelection_Method='cor',data=NUL
 #' @param label The label of dataset
 #' @param cutoff The cutoff used for feature selection threshold. It can be any value
 #' between 0 and 1.
-#' @param preMode The prediction mode. Available options are
-#' c('probability', 'classification').
+#' @param preMode The prediction mode. "Currently only supports 'probability' for binary classification tasks."
 #' @param classifier  Learners in mlr3
 #' @param cores The number of cores used for computation.
 #' @param verbose Whether to print running process information to the console
@@ -492,183 +494,6 @@ AddUnmapped=function(train=NULL,test=NULL,Unmapped_num=NULL,Add_FeartureSelectio
 }
 
 
-#' Selection of the optimal base model
-#'
-#' @param data The input training dataset. The first column
-#' is the label or the output. For binary classes,
-#' 0 and 1 are used to indicate the class member.
-#' @param pathlistDB A list of pathways with pathway IDs and their
-#' corresponding genes ('entrezID' is used).
-#' For details, please refer to ( data("GO2ALLEGS_BP") )
-#' @param FeatureAnno The annotation data stored in a data.frame for probe
-#' mapping. It must have at least two columns named 'ID' and 'entrezID'.
-#' (For details, please refer to data( data("MethylAnno") )
-#' @param resampling Resampling in mlr3verse.
-#' @param nfolds k-fold cross validation 
-#' @param classifiers A string of character vectors(Learners in mlr3)
-#' @param predMode The prediction mode. Available options are
-#' c('probability', 'classification').
-#' @param PathwaySizeUp The upper-bound of the number of genes in each
-#' biological pathways.
-#' @param PathwaySizeDown The lower-bound of the number of genes in each
-#' biological pathways.
-#' @param MinfeatureNum_pathways The minimal defined pathway size after mapping your
-#' own data to pathlistDB(KEGG database/GO database).
-#' @param Add_UnMapped Whether to add unmapped probes for prediction
-#' @param Unmapped_num The number of unmapped probes
-#' @param Add_FeartureSelection_Method Feature selection methods.
-#' @param Inner_CV Whether to perform a k-fold verification on the training set.
-#' @param inner_folds k-fold verification on the training set.
-#' @param Stage1_FeartureSelection_Method Feature selection methods.
-#' @param cutoff The cutoff used for feature selection threshold. It can be any value
-#' between 0 and 1.
-#' @param Stage2_FeartureSelection_Method Feature selection methods.
-#' @param cutoff2 The cutoff used for feature selection threshold. It can be any value
-#' between 0 and 1.
-#' @param cores The number of cores used for computation.
-#' @param verbose Whether to print running process information to the console
-#'
-#' @return A data frame containing the predictive performance of each basemodel
-#' @export
-#' @import ROCR
-#' @import caret
-
-HybaseModel=function(data=NULL,pathlistDB=NULL,FeatureAnno=NULL,resampling=NULL,nfolds=5,classifiers='liblinear', predMode = "probability",
-                     PathwaySizeUp=200,PathwaySizeDown=20,MinfeatureNum_pathways=10,
-                     Add_UnMapped=TRUE,Unmapped_num=300,Add_FeartureSelection_Method='wilcox.test',
-                     Inner_CV=TRUE,inner_folds=10,
-                     Stage1_FeartureSelection_Method='cor',cutoff=0.3,
-                     Stage2_FeartureSelection_Method='RemoveHighcor',cutoff2=0.95,
-                     cores=1,verbose=TRUE){
-  if(verbose)print('===================HybaseModel==================')
-  
-  if(Sys.info()[1]=="Windows"){
-    cores=1
-  }
-  
-  prediction=list()
-  FeatureAnno$ID=gsub('[\\.\\_\\-]','',FeatureAnno$ID)
-  colnames(TrainData)=gsub('[\\.\\_\\-]','',colnames(TrainData))
-  list_pathways=list()
-  Record=data.frame(resampling_id=1:nfolds,learner_name=1:nfolds,AUC=1:nfolds,ACC=1:nfolds,PCCs=1:nfolds)
-  Resampling=caret::createFolds(TrainData$label,k=nfolds)
-  if(!is.null(resampling)){
-    nfolds=resampling$param_set$values$folds
-  }
-  T1=Sys.time()
-  final=list()
-  for(lrn in 1:length(classifiers)){
-    classifier=classifiers[lrn]
-    if(verbose)print(paste0('<<<<<----- ',classifier,' ----->>>>>'))
-    for(xxx in 1:nfolds){
-      
-      sink(nullfile())
-      t1=Sys.time()
-      if(is.null(resampling)){
-        trainData=TrainData[unlist(Resampling[-xxx]),]
-        testData=TrainData[unlist(Resampling[xxx]),]
-      }else{
-        trainData=TrainData[resampling$train_set(xxx),]
-        testData=TrainData[resampling$test_set(xxx),]
-      }
-      geneNum_pathways=sapply(1:length(pathlistDB),function(i) length(pathlistDB[[i]]))
-      pathlistDB_sub=pathlistDB[which(geneNum_pathways > PathwaySizeDown & geneNum_pathways < PathwaySizeUp )]
-      featureAnno=FeatureAnno[FeatureAnno$ID %in% colnames(trainData),]
-      if(verbose)print(paste0('     |>Total number of pathways==>>',length(pathlistDB_sub)))
-      
-      
-      
-
-      feature_pathways=Stage1_FeartureSelection(Stage1_FeartureSelection_Method=Stage1_FeartureSelection_Method,data=trainData,cutoff=cutoff,
-                                                featureAnno=featureAnno,pathlistDB_sub=pathlistDB_sub,cores=cores,verbose=verbose)
-      
-      lens=sapply(1:length(feature_pathways),function(x) length(feature_pathways[[x]]))
-      
-
-      trainDataList=mclapply(1:length(feature_pathways),function(x) trainData[,feature_pathways[[x]]] ,mc.cores=cores)
-      testDataList=mclapply(1:length(feature_pathways),function(x) testData[,feature_pathways[[x]]] ,mc.cores=cores)
-      names(trainDataList)=names(pathlistDB_sub)
-      names(testDataList)=names(pathlistDB_sub)
-      trainDataList=trainDataList[which(lens>MinfeatureNum_pathways)]
-      testDataList=testDataList[which(lens>MinfeatureNum_pathways)]
-      featureNum_pathways=sapply(1:length(trainDataList),function(i2) length(trainDataList[[i2]]))
-      if(verbose)print(paste0('     |> Total number of selected pathways==>>',length(trainDataList)))
-      if(verbose)print(paste0('     |> Min features number of pathways==>>',min(featureNum_pathways)-1,'.......','Max features number of pathways==>>',max(featureNum_pathways)-1))
-      
-      
-
-      if(Inner_CV==TRUE){
-        if(verbose)print('     |> Using Inner CV ~ ~ ~')
-        train=mclapply(1:length(trainDataList),function(i4) baseModel(trainData =trainDataList[[i4]],testData =NULL,predMode =predMode,classifier = classifier,inner_folds=inner_folds),mc.cores=cores)
-        test=mclapply(1:length(testDataList),function(i5) baseModel(trainData =trainDataList[[i5]],testData =testDataList[[i5]],predMode =predMode,classifier = classifier),mc.cores=cores)
-      }else{
-        
-        train=mclapply(1:length(trainDataList),function(i4) baseModel(trainData =trainDataList[[i4]],testData =trainDataList[[i4]],predMode = predMode ,classifier = classifier),mc.cores=cores)
-        test=mclapply(1:length(testDataList),function(i5) baseModel(trainData =trainDataList[[i5]],testData =testDataList[[i5]],predMode = predMode ,classifier = classifier),mc.cores=cores)
-        
-        
-      }
-      
-
-      index=Stage2_FeartureSelection(Stage2_FeartureSelection_Method=Stage2_FeartureSelection_Method,data=train,
-                                     label=trainDataList[[1]]$label,cutoff=cutoff2,preMode='probability',classifier =classifier,verbose=verbose,cores=cores)
-      
-      
-      newtrain=do.call(cbind,train[index])
-      colnames(newtrain)=names(trainDataList)[index]
-      newtrain=cbind(label=trainDataList[[1]]$label,newtrain)
-      
-      newtest=do.call(cbind,test[index])
-      colnames(newtest)=names(trainDataList)[index]
-      newtest=cbind(label=testDataList[[1]]$label,newtest)
-      colnames(newtest)=gsub(':','',colnames(newtest))
-      colnames(newtrain)=gsub(':','',colnames(newtrain))
-      if(Add_UnMapped==TRUE){
-        Unmapped_Data=AddUnmapped(train=trainData,test=testData,Add_FeartureSelection_Method=Add_FeartureSelection_Method,
-                                  Unmapped_num=Unmapped_num,len=ncol(newtrain),anno=featureAnno,verbose=verbose,cores=cores)
-        newtrain=cbind(newtrain,Unmapped_Data$train)
-        newtest=cbind(newtest,Unmapped_Data$test)
-        if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
-      }
-      
-
-      
-      
-      classifier2=classifier
-      
-      result=baseModel(trainData=newtrain,testData=newtest,predMode ='probability',classifier = classifier2)
-      prediction_part=data.frame(sample=rownames(testData),prediction=result)
-      prediction[[xxx]]=prediction_part
-      names(prediction)[xxx]=paste('Resample No.',xxx)
-      testDataY=testDataList[[1]]$label
-      pre=ifelse(result>0.5,1,0)
-      Record[xxx,1]=xxx
-      Record[xxx,2]=classifier2
-      Record[xxx,5]=stats::cor(testDataY,result,method='pearson')
-      Record[xxx,3]=ROCR::performance(ROCR::prediction(result,testDataY),'auc')@y.values[[1]]
-      testDataY=as.factor(testDataY)
-      pre=as.factor(pre)
-      Record[xxx,4]=confusionMatrix(pre, testDataY)$overall['Accuracy'][[1]]
-   
-      sink()
-
-      if(xxx==nfolds){
-
-        T2=Sys.time()
-        if(verbose)print(paste0('{|>>>=====','Learner: ',classifier,'---Performance Metric---==>>','AUC:',round(mean(Record$AUC),digits = 3),' ','ACC:',round(mean(Record$ACC),digits = 3),' ','PCCs:',round(mean(Record$PCCs),digits = 3),'======<<<|}'))
-        final[[lrn]]=Record
-        if(verbose)print(T2-T1)
-        if(verbose)print('                       ')
-        
-      }
-    }
-  }
-  final=do.call(rbind,final)
-  final=aggregate(final[,3:5],by=list(learner_name=final$learner_name),mean)
-  final=final[order(final$AUC,decreasing = T),]
-  if(verbose)print(paste0('Best baseModel: ', final[1,1]))
-  return(final)
-}
 
 
 
@@ -689,8 +514,7 @@ HybaseModel=function(data=NULL,pathlistDB=NULL,FeatureAnno=NULL,resampling=NULL,
 #' @param resampling Resampling in mlr3verse.
 #' @param nfolds k-fold cross validation ( Only supported when TestData = NULL )
 #' @param classifier Learners in mlr3
-#' @param predMode The prediction mode. Available options are
-#' c('probability', 'classification').
+#' @param predMode The prediction mode. Currently only supports 'probability' for binary classification tasks.
 #' @param PathwaySizeUp The upper-bound of the number of genes in each
 #' biological pathways.
 #' @param PathwaySizeDown The lower-bound of the number of genes in each
@@ -799,7 +623,7 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
 
       if(verbose)print('Step2: FeartureSelection-features')
       feature_pathways=Stage1_FeartureSelection(Stage1_FeartureSelection_Method=Stage1_FeartureSelection_Method,data=trainData,cutoff=cutoff,
-                                               featureAnno=featureAnno,pathlistDB_sub=pathlistDB_sub,cores=cores,verbose=verbose)
+                                               featureAnno=featureAnno,pathlistDB_sub=pathlistDB_sub,MinfeatureNum_pathways=MinfeatureNum_pathways,cores=cores,verbose=verbose)
 
       lens=sapply(1:length(feature_pathways),function(x) length(feature_pathways[[x]]))
 
@@ -863,11 +687,15 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
         colnames(newtest)=gsub(':','',colnames(newtest))
         colnames(newtrain)=gsub(':','',colnames(newtrain))
         if(Add_UnMapped==TRUE){
-          Unmapped_Data=AddUnmapped(train=trainData,test=testData,Add_FeartureSelection_Method=Add_FeartureSelection_Method,
-                                    Unmapped_num=Unmapped_num,len=ncol(newtrain),anno=featureAnno,verbose=verbose,cores=cores)
-          newtrain=cbind(newtrain,Unmapped_Data$train)
-          newtest=cbind(newtest,Unmapped_Data$test)
-          if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+          if(Unmapped_num==0){
+            if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+          }else{
+            Unmapped_Data=AddUnmapped(train=trainData,test=testData,Add_FeartureSelection_Method=Add_FeartureSelection_Method,
+                                      Unmapped_num=Unmapped_num,len=ncol(newtrain),anno=featureAnno,verbose=verbose,cores=cores)
+            newtrain=cbind(newtrain,Unmapped_Data$train)
+            newtest=cbind(newtest,Unmapped_Data$test)
+            if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+          }
         }
 
         #(Predict and Metric)
@@ -883,13 +711,13 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
         testDataY=testDataList[[1]]$label
         pre=ifelse(result>0.5,1,0)
         Record[xxx,1]=xxx
-        Record[xxx,2]=classifier2
+        Record[xxx,2]=ifelse(is.character(classifier),classifier,classifier$id)
         Record[xxx,5]=stats::cor(testDataY,result,method='pearson')
         Record[xxx,3]=ROCR::performance(ROCR::prediction(result,testDataY),'auc')@y.values[[1]]
-        testDataY=as.factor(testDataY)
-        pre=as.factor(pre)
-        Record[xxx,4]=confusionMatrix(pre, testDataY)$overall['Accuracy'][[1]]
-        if(verbose)print(paste0('######Resampling NO.',xxx,'~~~~',classifier2,'==>','AUC:',round(Record[xxx,3],digits = 3),' ','ACC:',round(Record[xxx,4],digits = 3),' ','PCCs:',round(Record[xxx,5],digits = 3)))
+        accuracy_class1 <- sum(pre[testDataY == 1] == 1) / sum(testDataY == 1)
+        accuracy_class0 <- sum(pre[testDataY == 0] == 0) / sum(testDataY == 0)
+        Record[xxx,4]=(accuracy_class1 + accuracy_class0) / 2
+        if(verbose)print(paste0('######Resampling NO.',xxx,'~~~~',ifelse(is.character(classifier),classifier,classifier$id),'==>','AUC:',round(Record[xxx,3],digits = 3),' ','BAC:',round(Record[xxx,4],digits = 3),' ','PCCs:',round(Record[xxx,5],digits = 3)))
         t2=Sys.time()
         if(verbose)print(t2-t1)
         if(verbose)print('---------------------####################------------------')
@@ -958,7 +786,7 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
 
     if(verbose)print('Step2: FeartureSelection-features')
     feature_pathways=Stage1_FeartureSelection(Stage1_FeartureSelection_Method=Stage1_FeartureSelection_Method,data=trainData,cutoff=cutoff,
-                                             featureAnno=featureAnno,pathlistDB_sub=pathlistDB_sub,cores=cores,verbose=verbose)
+                                             featureAnno=featureAnno,pathlistDB_sub=pathlistDB_sub,MinfeatureNum_pathways=MinfeatureNum_pathways,cores=cores,verbose=verbose)
 
     lens=sapply(1:length(feature_pathways),function(x) length(feature_pathways[[x]]))
 
@@ -1049,11 +877,15 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
       colnames(newtest)=gsub(':','',colnames(newtest))
       colnames(newtrain)=gsub(':','',colnames(newtrain))
       if(Add_UnMapped==TRUE){
-        Unmapped_Data=AddUnmapped(train=trainData,test=testData,,Add_FeartureSelection_Method=Add_FeartureSelection_Method,
-                                  Unmapped_num=Unmapped_num,anno=featureAnno,verbose=verbose,cores=cores)
-        newtrain=cbind(newtrain,Unmapped_Data$train)
-        newtest=cbind(newtest,Unmapped_Data$test)
-        if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+        if(Unmapped_num==0){
+          if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+        }else{
+          Unmapped_Data=AddUnmapped(train=trainData,test=testData,Add_FeartureSelection_Method=Add_FeartureSelection_Method,
+                                    Unmapped_num=Unmapped_num,len=ncol(newtrain),anno=featureAnno,verbose=verbose,cores=cores)
+          newtrain=cbind(newtrain,Unmapped_Data$train)
+          newtest=cbind(newtest,Unmapped_Data$test)
+          if(verbose)print(paste0('     |> Merge PathwayFeature and AddFeature ==>>',ncol(newtrain)))
+        }
       }
 
       #(Predict and Metric)
@@ -1065,13 +897,13 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
       predict=data.frame(sample=rownames(testData),prediction=result)
       testDataY=testDataList[[1]]$label
       pre=ifelse(result>0.5,1,0)
-      Record[1,1]=classifier
+      Record[1,1]=ifelse(is.character(classifier),classifier,classifier$id)
       Record[1,4]=stats::cor(testDataY,result,method='pearson')
       Record[1,2]=ROCR::performance(ROCR::prediction(result,testDataY),'auc')@y.values[[1]]
-      testDataY=as.factor(testDataY)
-      pre=as.factor(pre)
-      Record[1,3]=confusionMatrix(pre, testDataY)$overall['Accuracy'][[1]]
-      if(verbose)print(paste0('######~~~~',classifier2,'==>','AUC:',round(Record[1,2],digits = 3),' ','ACC:',round(Record[1,3],digits = 3),' ','PCCs:',round(Record[1,4],digits = 3)))
+      accuracy_class1 <- sum(pre[testDataY == 1] == 1) / sum(testDataY == 1)
+      accuracy_class0 <- sum(pre[testDataY == 0] == 0) / sum(testDataY == 0)
+      Record[1,3]=(accuracy_class1 + accuracy_class0) / 2
+      if(verbose)print(paste0('######~~~~',ifelse(is.character(classifier),classifier,classifier$id),'==>','AUC:',round(Record[1,2],digits = 3),' ','BAC:',round(Record[1,3],digits = 3),' ','PCCs:',round(Record[1,4],digits = 3)))
       final=list('Prediction'=predict,'Metric'=Record)
       T2=Sys.time()
       if(verbose)print(T2-T1)
@@ -1083,6 +915,281 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
 
 
 
+
+#' BioM2 Hyperparametric Combination
+#'
+#' @param TrainData The input training dataset. The first column
+#' is the label or the output. For binary classes,
+#' 0 and 1 are used to indicate the class member.
+#' @param pathlistDB A list of pathways with pathway IDs and their
+#' corresponding genes ('entrezID' is used).
+#' For details, please refer to ( data("GO2ALLEGS_BP") )
+#' @param FeatureAnno The annotation data stored in a data.frame for probe
+#' mapping. It must have at least two columns named 'ID' and 'entrezID'.
+#' (For details, please refer to data( data("MethylAnno") )
+#' @param resampling Resampling in mlr3verse.
+#' @param nfolds k-fold cross validation ( Only supported when TestData = NULL )
+#' @param classifier Learners in mlr3
+#' @param predMode The prediction mode. Currently only supports 'probability' for binary classification tasks.
+#' @param PathwaySizeUp The upper-bound of the number of genes in each
+#' biological pathways.
+#' @param PathwaySizeDown The lower-bound of the number of genes in each
+#' biological pathways.
+#' @param MinfeatureNum_pathways The minimal defined pathway size after mapping your
+#' own data to pathlistDB(KEGG database/GO database).
+#' @param Add_UnMapped Whether to add unmapped probes for prediction
+#' @param Unmapped_num The number of unmapped probes
+#' @param Add_FeartureSelection_Method Feature selection methods.
+#' @param Inner_CV Whether to perform a k-fold verification on the training set.
+#' @param inner_folds k-fold verification on the training set.
+#' @param Stage1_FeartureSelection_Method Feature selection methods.
+#' @param stage1_cutoff The cutoff used for feature selection threshold. It can be any value
+#' between 0 and 1.
+#' @param Stage2_FeartureSelection_Method Feature selection methods.
+#' @param stage2_cutoff The cutoff used for feature selection threshold. It can be any value
+#' between 0 and 1.
+#' @param classifier2 Learner for stage 2 prediction(if classifier2==NULL,then it is the same as the learner in stage 1.)
+#' @param cores The number of cores used for computation.
+#' @param verbose Whether to print running process information to the console
+#'
+#'
+#' @return A data frame contains hyperparameter results
+#' @export
+#' @import ROCR
+#' @import caret
+#' @importFrom utils head 
+#' @importFrom stats  wilcox.test 
+
+#'
+HyBioM2=function(TrainData=NULL,pathlistDB=NULL,FeatureAnno=NULL,resampling=NULL,nfolds=5,classifier='liblinear', predMode = "probability",
+                 PathwaySizeUp=200,PathwaySizeDown=20,MinfeatureNum_pathways=10,
+                 Add_UnMapped=TRUE,Add_FeartureSelection_Method='wilcox.test',Unmapped_num=300,
+                 Inner_CV=TRUE,inner_folds=10,
+                 Stage1_FeartureSelection_Method='cor',stage1_cutoff=0.3,
+                 Stage2_FeartureSelection_Method='RemoveHighcor',stage2_cutoff=0.8,
+                 classifier2=NULL,cores=1,verbose=TRUE){
+  re=list()
+  if(Sys.info()[1]=="Windows"){
+    cores=1
+  }
+  if(verbose)print('===================HyBioM2==================')
+  for(c1 in 1:length(classifier)){
+    stage1_learner=classifier[[c1]]
+    HOPE=list()
+    t1=Sys.time()
+    for(luck in 1:length(stage1_cutoff)){
+      set.seed(666)
+      cutoff=stage1_cutoff[luck]
+      Resampling=createFolds(TrainData$label,k=nfolds)
+      final=list()
+      
+      pr=sum(TrainData$label==1)>sum(TrainData$label==0)
+      for(xxx in 1:nfolds){
+        #print('Step1: ReadData')
+        trainData=TrainData[unlist(Resampling[-xxx]),]
+        testData=TrainData[unlist(Resampling[xxx]),]
+        geneNum_pathways=sapply(1:length(pathlistDB),function(i) length(pathlistDB[[i]]))
+        pathlistDB_sub=pathlistDB[which(geneNum_pathways > PathwaySizeDown & geneNum_pathways < PathwaySizeUp )]
+        #print(paste0('     |>Total number of pathways==>>',length(pathlistDB_sub)))
+        
+        
+        #print('Step2: FeartureSelection-features')
+        if(Stage1_FeartureSelection_Method=='cor'){
+          #print(paste0('      Using <<  correlation  >>',' ,and you choose cutoff:',cutoff))
+          Cor=stats::cor(trainData$label,trainData)
+          Cor=ifelse(Cor>0,Cor,-Cor)
+          names(Cor)=colnames(trainData)
+          Cor_names=names(Cor)
+          Cor_cutoff=Cor[which(Cor>cutoff)]
+          Cor_cutoff_names=names(Cor_cutoff)
+        }else{
+          #print(paste0('      Using <<  wilcox.test  >>',' ,and you choose cutoff:',cutoff))
+          train_0=trainData[which(trainData$label==0),]
+          train_1=trainData[which(trainData$label==1),]
+          Cor=unlist(mclapply(1:ncol(trainData),function(x) wilcox.test(train_0[,x],train_1[,x])$p.value,mc.cores=cores))
+          
+          names(Cor)=colnames(trainData)
+          Cor_names=names(Cor)
+          Cor_cutoff=Cor[which(Cor<cutoff)]
+          Cor_cutoff_names=names(Cor_cutoff)
+        }
+        
+        featureAnno=FeatureAnno[FeatureAnno$ID %in% colnames(trainData),]
+        MinfeatureNum_pathways2=MinfeatureNum_pathways+1
+        featureNum_pathways=mclapply(1:length(pathlistDB_sub),function(x){
+          id=c('label',featureAnno$ID[which(featureAnno$entrezID %in% pathlistDB_sub[[x]])])
+          if(length(id)>MinfeatureNum_pathways2){
+            id2=id[which(id %in% Cor_cutoff_names)]
+            if(length(id2)<MinfeatureNum_pathways2){
+              a=Cor[id]
+              if(Stage1_FeartureSelection_Method=='cor'){
+                id2=names(a)[order(a,decreasing = T)[1:MinfeatureNum_pathways2]]
+              }else{
+                id2=names(a)[order(a,decreasing = F)[1:MinfeatureNum_pathways2]]
+              }
+              return(id2)
+            }else{
+              return(id2)
+            }
+          }else{
+            return(id)
+          }
+        } ,mc.cores=cores)
+        lens=sapply(1:length(featureNum_pathways),function(x) length(featureNum_pathways[[x]]))
+        
+        #print('Step3: MergeData')
+        trainDataList=mclapply(1:length(featureNum_pathways),function(x) trainData[,featureNum_pathways[[x]]] ,mc.cores=cores)
+        testDataList=mclapply(1:length(featureNum_pathways),function(x) testData[,featureNum_pathways[[x]]] ,mc.cores=cores)
+        names(trainDataList)=names(pathlistDB_sub)
+        names(testDataList)=names(pathlistDB_sub)
+        trainDataList=trainDataList[which(lens>MinfeatureNum_pathways)]
+        testDataList=testDataList[which(lens>MinfeatureNum_pathways)]
+        featureNum_pathways2=sapply(1:length(trainDataList),function(i2) length(trainDataList[[i2]]))
+        #print(paste0('     |> Total number of selected pathways==>>',length(trainDataList)))
+        #print(paste0('     |> Min features number of pathways==>>',min(featureNum_pathways2)-1,'.......','Max features number of pathways==>>',max(featureNum_pathways2)-1))
+        
+        
+        #(FeartureSelection-pathways)
+        #print('Step4: Reconstruction')
+        if(Inner_CV){
+          #print('     |> Using Inner CV ~ ~ ~')
+          train=mclapply(1:length(trainDataList),function(i4) baseModel(trainData =trainDataList[[i4]],testData =NULL,predMode =predMode,classifier = classifier,inner_folds=inner_folds),mc.cores=cores)
+          test=mclapply(1:length(testDataList),function(i5) baseModel(trainData =trainDataList[[i5]],testData =testDataList[[i5]],predMode =predMode,classifier = classifier),mc.cores=cores)
+          #pred=mclapply(1:length(trainDataList),function(i4) baseModel2(trainData =trainDataList[[i4]],testData =testDataList[[i4]],classifier = stage1_learner,inner_folds=inner_folds),mc.cores=cores)
+          #train=lapply(1:length(pred),function(x) pred[[x]]$predtrain)
+          #test=lapply(1:length(pred),function(x) pred[[x]]$predtest)
+        }else{
+          
+          train=mclapply(1:length(trainDataList),function(i4) baseModel(trainData =trainDataList[[i4]],testData =trainDataList[[i4]],predMode = predMode ,classifier = stage1_learner),mc.cores=cores)
+          test=mclapply(1:length(testDataList),function(i5) baseModel(trainData =trainDataList[[i5]],testData =testDataList[[i5]],predMode = predMode ,classifier = stage1_learner),mc.cores=cores)
+          
+        }
+        
+        #print('     <<< Reconstruction Done! >>>     ')
+        
+        #(FeartureSelection-pathways)
+        #print('Step5: FeartureSelection-pathways')
+        
+        
+        dbmap=unique(unlist(pathlistDB_sub))
+        annomap=unique(featureAnno$entrezID)
+        mapgene=intersect(annomap,dbmap)
+        map=featureAnno$ID[which(featureAnno$entrezID %in% mapgene)]
+        Unmapped_Train=trainData[,setdiff(colnames(trainData),map)]
+        Unmapped_Test=testData[,setdiff(colnames(trainData),map)]
+        #print(paste0('Unmapped_num: ',ncol(Unmapped_Train)))
+        Unmapped_0=Unmapped_Train[which(Unmapped_Train$label==unique(Unmapped_Train$label)[1]),]
+        Unmapped_1=Unmapped_Train[which(Unmapped_Train$label==unique(Unmapped_Train$label)[2]),]
+        if(Add_FeartureSelection_Method=='cor'){
+          Unmapped_pvalue= abs(stats::cor(Unmapped_Train$label,Unmapped_Train[,-1]))
+          Unmapped_pvalue=1/Unmapped_pvalue
+        }else{
+          Unmapped_0=Unmapped_Train[which(Unmapped_Train$label==unique(Unmapped_Train$label)[1]),]
+          Unmapped_1=Unmapped_Train[which(Unmapped_Train$label==unique(Unmapped_Train$label)[2]),]
+          Unmapped_pvalue=unlist(mclapply(2:ncol(Unmapped_Train),function(x) wilcox.test(Unmapped_0[,x],Unmapped_1[,x])$p.value,mc.cores=cores))
+          
+        }
+        
+        Record2=list()
+        for(ii in  1:length(stage2_cutoff)){
+          
+          if(stage2_cutoff[ii]==0){
+            index=Stage2_FeartureSelection(Stage2_FeartureSelection_Method='None',data=train,
+                                           label=trainDataList[[1]]$label,cutoff=stage2_cutoff[ii],preMode='probability',classifier =classifier,verbose = FALSE,cores=cores)
+          }else{
+            index=Stage2_FeartureSelection(Stage2_FeartureSelection_Method=Stage2_FeartureSelection_Method,data=train,
+                                           label=trainDataList[[1]]$label,cutoff=stage2_cutoff[ii],preMode='probability',classifier =classifier,verbose = FALSE,cores=cores)
+            
+          }
+          
+          if(is.null(classifier2)){
+            stage2_learners=stage1_learner
+          }else{
+            stage2_learners=classifier2
+          }
+          
+          n=length(Unmapped_num)*length(stage2_learners)
+          Record=data.frame(Unmapped_num=1:n,stage2_learner=1:n,AUC=1:n,
+                            PCC=1:n,BAC=1:n,cutoff=1:n)
+          for(i in 1:length(Unmapped_num)){
+            newtrain=do.call(cbind,train[index])
+            colnames(newtrain)=names(trainDataList)[index]
+            newtrain=cbind(label=trainDataList[[1]]$label,newtrain)
+            
+            newtest=do.call(cbind,test[index])
+            colnames(newtest)=names(trainDataList)[index]
+            newtest=cbind(label=testDataList[[1]]$label,newtest)
+            colnames(newtest)=gsub(':','',colnames(newtest))
+            colnames(newtrain)=gsub(':','',colnames(newtrain))
+            if(Add_UnMapped==TRUE & Unmapped_num[i] > 0 ){
+              if(length(Unmapped_pvalue)<Unmapped_num[i]){
+                Unmapped_Num=length(Unmapped_pvalue)
+              }else{
+                Unmapped_Num=Unmapped_num[i]
+              }
+              Unmapped_id=order(Unmapped_pvalue)[1:Unmapped_Num]
+              Unmapped_id=Unmapped_id+1
+              Unmapped_Train2=Unmapped_Train[,Unmapped_id]
+              Unmapped_Test2=Unmapped_Test[,Unmapped_id]
+              newtrain=cbind(newtrain,Unmapped_Train2)
+              newtest=cbind(newtest,Unmapped_Test2)
+            }
+            
+            
+            for(z in 1:length(stage2_learners)){
+              classifier_2=stage2_learners[[z]]
+              a=(i-1)*length(stage2_learners)+z
+              result=baseModel(trainData=newtrain,testData=newtest,predMode ='probability',classifier = classifier_2)
+              testDataY=testDataList[[1]]$label
+              pre=ifelse(result>0.5,1,0)
+              accuracy_class1 <- sum(pre[testDataY == 1] == 1) / sum(testDataY == 1)
+              accuracy_class0 <- sum(pre[testDataY == 0] == 0) / sum(testDataY == 0)
+              Record[a,5]=(accuracy_class1 + accuracy_class0) / 2
+              Record[a,1]=Unmapped_num[i]
+              Record[a,2]=ifelse(is.character(classifier_2),classifier_2,classifier_2$id)
+              Record[a,4]=stats::cor(testDataY,result,method='pearson')
+              Record[a,3]=ROCR::performance(ROCR::prediction(result,testDataY),'auc')@y.values[[1]]
+              Record[a,6]=stage2_cutoff[ii]
+            }
+          }
+          Record2[[ii]]=Record
+          
+          
+          
+        }
+
+        Record2=do.call(rbind,Record2) 
+        final[[xxx]]=Record2
+      }
+      
+      
+      hope=do.call(rbind,final)
+      
+      hope=aggregate(hope[,c(3:5)],by=list(Unmapped_num=hope$Unmapped_num,stage2_cutoff=hope$cutoff,stage2_learner=hope$stage2_learner),mean)
+      hope$stage1_cutoff=stage1_cutoff[luck]
+      #if(verbose)print(hope)
+      #print(paste0('stage1_cutoff : ',cutoff))
+      #print(luck)
+      
+      HOPE[[luck]]=hope
+    }
+    HOPE=do.call(rbind,HOPE)
+    HOPE$stage1_learner=ifelse(is.character(stage1_learner),stage1_learner,stage1_learner$id)
+    HOPE=HOPE[,c('stage1_learner','stage2_learner','stage1_cutoff','stage2_cutoff','Unmapped_num','AUC','BAC','PCC')]
+    if(verbose)print(HOPE)
+    re[[c1]]=HOPE
+    t2=Sys.time()
+    if(verbose)print(t2-t1)
+    if(verbose)print(' ')
+  }
+  re=do.call(rbind,re)
+  re=re[order(re$AUC,decreasing = T),]
+  return(re)
+}
+
+
+
+
 #' Find suitable parameters for partitioning pathways modules
 #'
 #' @param pathways_matrix A pathway matrix generated by the BioM2( target='pathways') function.
@@ -1104,7 +1211,7 @@ BioM2=function(TrainData=NULL,TestData=NULL,pathlistDB=NULL,FeatureAnno=NULL,res
 #'
 #'
 #'
-FindParaModule=function(pathways_matrix=NULL,control_label=NULL,minModuleSize = seq(10,20,5),mergeCutHeight=seq(0,0.3,0.1),minModuleNum=20,power=NULL,exact=TRUE,ancestor_anno=NULL){
+FindParaModule=function(pathways_matrix=NULL,control_label=0,minModuleSize = seq(10,20,5),mergeCutHeight=seq(0,0.3,0.1),minModuleNum=5,power=NULL,exact=TRUE,ancestor_anno=NULL){
   if('package:WGCNA' %in% search()){
     final=list()
     if(exact==FALSE & is.null(ancestor_anno)){
@@ -1125,11 +1232,9 @@ FindParaModule=function(pathways_matrix=NULL,control_label=NULL,minModuleSize =
 
     if(is.null(power)){
       powers = c(c(1:10), seq(from = 12, to=20, by=1))
-      sink(nullfile())
-      sft = pickSoftThreshold(data, powerVector = powers, verbose = 5)
-      sink()
+      sft = pickSoftThreshold(data, powerVector = powers, verbose = 0)
       if(is.na(sft$powerEstimate)){
-        stop('Could not find a proper powers , Please give a power by youself .')
+        return('Could not find a proper powers , Please give a power by youself .')
       }else{
         power=sft$powerEstimate
         message('Find the proper power!')
@@ -1146,15 +1251,13 @@ FindParaModule=function(pathways_matrix=NULL,control_label=NULL,minModuleSize =
         n=length(cutoff)
         result=data.frame(mergeCutHeight=1:n,Number_clusters=1:n,Mean_number_pathways=1:n,Mean_Fraction=1:n,Sd_Fraction=1:n,minModuleSize=1:n)
         for(ii in 1:length(cutoff)){
-          sink(nullfile())
           net = blockwiseModules(data, power = power,
                                  TOMType = "unsigned", minModuleSize = Num_module[xxx],
                                  reassignThreshold = 0, mergeCutHeight = cutoff[ii],
                                  numericLabels = TRUE, pamRespectsDendro = FALSE,
                                  saveTOMs = F,
                                  saveTOMFileBase = "femaleMouseTOM",
-                                 verbose = 3)
-          sink()
+                                 verbose = 0)
           cluster=data.frame(ID=names(net$colors),cluster=net$colors)
           cluster$cluster=cluster$cluster+1
           cluster_list=list()
@@ -1255,24 +1358,20 @@ PathwaysModule=function(pathways_matrix=NULL,control_label=NULL,power=NULL,minMo
 
     if(is.null(power)){
       powers = c(1:30)
-      sink(nullfile())
-      sft = pickSoftThreshold(data, powerVector = powers, verbose = 5)
-      sink()
+      sft = pickSoftThreshold(data, powerVector = powers, verbose = 0)
       if(is.na(sft$powerEstimate)){
-        stop('Could not find a proper powers , Please give a power by youself .')
+        return('Could not find a proper powers , Please give a power by youself .')
       }else{
         power=sft$powerEstimate
         message('Find the proper power!')
       }
     }
-    sink(nullfile())
     net = blockwiseModules(data, power = power,
                            TOMType = "unsigned", minModuleSize = minModuleSize,
                            reassignThreshold = 0, mergeCutHeight = mergeCutHeight,
                            numericLabels = TRUE, pamRespectsDendro = FALSE,
                            saveTOMs = F,
-                           verbose = 3)
-    sink()
+                           verbose = 0)
     cluster=data.frame(ID=names(net$colors),cluster=net$colors)
     cluster_list=list()
     faction=vector()
